@@ -4,14 +4,19 @@ import re
 from typing import Dict, Optional, Tuple, Union
 
 import numpy as np
-from numpy import array, ascontiguousarray, zeros, zeros_like
+from numpy import array, float64, zeros, zeros_like
 import pandas as pd
 from scipy.special import expit as sigmoid
 from sklearn.base import BaseEstimator
 from sklearn.metrics import mean_squared_error
 
 from harmoniums.const import Matrix, MatrixOrTriplet, MatrixTriplet, Observation
-from harmoniums.utils import generate_binary_permutations, MiniBatchIterator
+from harmoniums.utils import (
+    generate_binary_permutations,
+    get_columns,
+    reset_random_state,
+    MiniBatchIterator,
+)
 
 
 class BaseHarmonium(BaseEstimator, ABC):
@@ -42,6 +47,17 @@ class BaseHarmonium(BaseEstimator, ABC):
         """
         return {param: getattr(self, param, None) for param in self.parameters}
 
+    def set_parameters(self, values: dict):
+        """
+        Set free parameter values by name.
+        """
+        for name, value in values.items():
+            if name not in self.parameters:
+                raise KeyError(
+                    f"Parameter {name} is not a free parameter of this model."
+                )
+            setattr(self, name, value)
+
     @property
     @abstractmethod
     def no_penalty(self) -> Tuple[str, ...]:
@@ -105,7 +121,7 @@ class BaseHarmonium(BaseEstimator, ABC):
         self.X_validation = X_validation
 
         # Set random state.
-        np.random.seed(random_state)
+        reset_random_state(random_state)
 
         self.is_parameters_initialised_ = False
         # Previous state (empty), when using persistent contrastive divergence.
@@ -120,7 +136,7 @@ class BaseHarmonium(BaseEstimator, ABC):
             return True
         return False
 
-    def _unpack_settings_as_tuple(self, param_settings) -> tuple:
+    def _unpack_settings(self, param_settings) -> tuple:
         """
         Unpack parameter specific settings as a `self.parameters` like tuple.
         """
@@ -141,7 +157,7 @@ class BaseHarmonium(BaseEstimator, ABC):
             learning_rate = tuple(self.learning_rate for _ in self.parameters)
         else:
             try:
-                learning_rate = self._unpack_settings_as_tuple(self.learning_rate)
+                learning_rate = self._unpack_settings(self.learning_rate)
             except TypeError:
                 raise ValueError("Incorrect learning rate setting.")
         return learning_rate
@@ -158,13 +174,13 @@ class BaseHarmonium(BaseEstimator, ABC):
             )
         else:
             try:
-                weight_decay = self._unpack_settings_as_tuple(self.weight_decay)
+                weight_decay = self._unpack_settings(self.weight_decay)
             except TypeError:
                 raise ValueError("Incorrect weight decay setting.")
         return weight_decay
 
     def _copy(self, X):
-        """ Return a copy if `x` is not None. """
+        """Return a copy if `x` is not None."""
         if X is None:
             raise ValueError("None in _copy")
             return None
@@ -179,18 +195,6 @@ class BaseHarmonium(BaseEstimator, ABC):
         """
         return (self._copy(X[0]), self._copy(X[1]), self._copy(X[2]))
 
-    def _get_columns(self, X: Matrix, columns: list) -> Matrix:
-        """
-        Unpack the columns from the matrix.
-        """
-        if columns:
-            # Unpack data frame as Numpy matrices.
-            if isinstance(X, pd.DataFrame):
-                return ascontiguousarray(X[columns].values)
-            else:
-                return X[:, columns]
-        return array([])
-
     def _set_columns(self, X: Matrix, X_values: Matrix, columns: list) -> Matrix:
         """
         Set specific values of matrix.
@@ -369,7 +373,8 @@ class BaseHarmonium(BaseEstimator, ABC):
         with np.printoptions(precision=4, floatmode="fixed"):
             print(
                 column_format.format(
-                    self.mini_batcher_iterator_.current_epoch, *latest_metrics_combined,
+                    self.mini_batcher_iterator_.current_epoch,
+                    *latest_metrics_combined,
                 )
             )
 
@@ -377,21 +382,22 @@ class BaseHarmonium(BaseEstimator, ABC):
         """
         Are the parameters converged or is the maximum iteration number exceeded?
         """
-        # Check the size of all parameter updates.
-        parameters_converged = map(
-            lambda x: np.all(abs(x) < self.tolerance), parameter_updates
-        )
-        if np.all(np.fromiter(parameters_converged, dtype=bool)):
-            if self.verbose:
-                print("Iteration converged!")
-            return True
-        elif (
-            self.maximum_iteration > 0 and self.iteration_step >= self.maximum_iteration
-        ):
+        if self.maximum_iteration > 0 and self.iteration_step >= self.maximum_iteration:
             if self.verbose:
                 print("Terminated, exceeded maximum number of iterations.")
             return True
 
+        # Check convergence of parameters every 100 iterations, using the size
+        # of all parameter updates.
+        if self.iteration_step % 100 == 0:
+            parameters_converged = map(
+                lambda x: np.all(abs(x) < self.tolerance), parameter_updates
+            )
+            if np.fromiter(parameters_converged, dtype=bool).all():
+                if self.verbose:
+                    print("Iteration converged!")
+                return True
+
         return False
 
     def _unpack(self, X: Matrix) -> Tuple[Matrix, Matrix]:
@@ -399,8 +405,8 @@ class BaseHarmonium(BaseEstimator, ABC):
         Unpack variables from `X`.
         """
         return (
-            self._get_columns(X, self.visible_columns_),
-            self._get_columns(X, self.event_columns),
+            get_columns(X, self.visible_columns_),
+            get_columns(X, self.event_columns),
         )
 
     def _compress(self, xi: Matrix, event: Matrix) -> Matrix:
@@ -451,7 +457,7 @@ class BaseHarmonium(BaseEstimator, ABC):
         U = np.random.uniform(size=(m, self.n_hidden_units))
 
         # Turn hidden unit on when probability is larger than random uniform number,
-        H = (P > U).astype(int)
+        H = (P > U).astype(float64)
         return H
 
     def transform(
