@@ -43,7 +43,9 @@ from numpy import (
     random,
     sqrt,
     sum,
+    take,
     unique,
+    where,
     zeros,
     zeros_like,
 )
@@ -58,6 +60,7 @@ from harmoniums.const import Matrix, MatrixPair, MatrixTriplet, ObservationMatri
 from harmoniums.const import ObservationTriplet as Observation
 from harmoniums.distributions import (
     fit_truncated_gamma_parameters,
+    log_normalisation_gamma_distribution,
     normalisation_gamma_interval_distribution,
     normalisation_gamma_distribution,
 )
@@ -75,7 +78,9 @@ from harmoniums.utils import (
     brier_loss,
     check_arrays,
     generate_binary_permutations,
+    get_columns,
     hash_array,
+    reset_random_state,
 )
 
 
@@ -203,7 +208,7 @@ class SurvivalHarmonium(BaseHarmonium):
         if copy:
             X = X.copy()
 
-        XI_B = self._get_columns(X, self.survival_columns)
+        XI_B = get_columns(X, self.survival_columns)
         # Normalise to [0, 1] range.
         time_horizon = getattr(self, "time_horizon_", 1.0)
 
@@ -223,21 +228,20 @@ class SurvivalHarmonium(BaseHarmonium):
         if normalise and self.n_event_units > 0:
             X = self._normalise(X)
 
-        X_A = self._get_columns(X, self.categorical_columns)
-        X_B = self._get_columns(X, self.survival_columns)
-        X_C = self._get_columns(X, self.numeric_columns)
+        X_A = get_columns(X, self.categorical_columns)
+        X_B = get_columns(X, self.survival_columns)
+        X_C = get_columns(X, self.numeric_columns)
 
         # Reshape empty arrays to size m x 0.
         m = self.get_number_of_rows((X_A, X_B, X_C))
-        XI = (X_A.astype(float).reshape(m, -1), X_B.reshape(m, -1), X_C.reshape(m, -1))
+        XI = (X_A.reshape(m, -1), X_B.reshape(m, -1), X_C.reshape(m, -1))
 
-        event = self._get_columns(X, self.event_columns)
+        event = get_columns(X, self.event_columns, dtype=bool_)
 
         # Check that the data admits to the bounds.
         if normalise and verify and self.n_event_units > 0:
             if event.size == 0:
-                event = ones_like(X_B)
-            event = event.astype(bool)
+                event = ones_like(X_B, order="C", dtype=bool_)
             # Time-to-event variables must be strictly > 0 when observed,
             # or >= 0 when censored.
             assert all((X_B > 0) | ((X_B >= 0) & ~event))
@@ -272,35 +276,6 @@ class SurvivalHarmonium(BaseHarmonium):
 
         return X
 
-    def fill_nan(self, XI: MatrixTriplet) -> MatrixTriplet:
-        """
-        Fill all nan values using precomputed values according to `fill_nan_method`.
-
-        Returns (MatrixTriplet): Copy of the original matrix.
-        """
-        filled_triplet: List[Matrix] = []
-        for i, columns in enumerate(
-            [self.categorical_columns, self.survival_columns, self.numeric_columns]
-        ):
-            if (xi_i_new := self._copy(XI[i])) is not None:
-                nan_inds = np.where(isnan(xi_i_new))
-                if xi_i_new[nan_inds].size > 0:
-                    # For the categorical (group A) and numerical data (group
-                    # C), fill using values calculating with the
-                    # `fill_nan_method`.
-                    if i != 1:
-                        fillers = self._get_columns(self.fill_nan_, columns)
-                        xi_i_new[nan_inds] = np.take(fillers, nan_inds[1])
-                    # For the time-to-event variables (group B), replace nan->0.
-                    # This ensures that `sample_x_wake` generates samples from
-                    # the entire [0,1] interval (instead of [xi, 1]).
-                    else:
-                        xi_i_new[nan_inds] = 0.0
-                filled_triplet.append(xi_i_new)
-            else:
-                filled_triplet.append(None)
-        return (filled_triplet[0], filled_triplet[1], filled_triplet[2])
-
     def initialise_parameters(self, X: Optional[Matrix] = None):
         """
         Initialise training parameteres, possibly using training data `X`.
@@ -323,11 +298,15 @@ class SurvivalHarmonium(BaseHarmonium):
         )
         self.a_A = zeros(shape=(self.n_categorical_units, 1), **numpy_args)
         if X is not None and self.n_categorical_units > 0:
-            (XI_A, _, _,), _ = self._unpack(X, normalise=True)
+            (
+                XI_A,
+                _,
+                _,
+            ), _ = self._unpack(X, normalise=True)
             # Initialise the bias using probabilities.
             p = nanmean(XI_A, axis=0, keepdims=True).T
             # Set visible units according to ln[p/(1-p)], see Ref. [1].
-            self.a_A = log(p / (1 - p), out=zeros_like(p), where=(p != 0))
+            self.a_A = log(p / (1 - p), out=zeros_like(p), where=(p != 0) & (p != 1))
 
         # 2) Initialise time-to-event parameters (group B).
         # Glorot-Bengio weight initialisations.
@@ -344,7 +323,11 @@ class SurvivalHarmonium(BaseHarmonium):
         self.a_B = zeros((self.n_event_units, 1), **numpy_args)
         # Multiply GB scale by 2 to account for asymmetry.
         if X is not None and self.n_event_units > 0:
-            (_, XI_B, _,), _ = self._unpack(X, normalise=True)
+            (
+                _,
+                XI_B,
+                _,
+            ), _ = self._unpack(X, normalise=True)
             # Expect bias when half the latent states turn on.
             Vh_exp = sum(self.V, axis=1, keepdims=True) / 2
 
@@ -373,7 +356,9 @@ class SurvivalHarmonium(BaseHarmonium):
 
         # 3) Initialise parameters of numerical features (group C).
         self.W_C = random.uniform(
-            -scale_GB, scale_GB, size=(self.n_numeric_units, self.n_hidden_units),
+            -scale_GB,
+            scale_GB,
+            size=(self.n_numeric_units, self.n_hidden_units),
         )
         self.sigma = ones(shape=(self.n_numeric_units, 1), **numpy_args)
         # TODO [#40]: Use more clever initialisation for bias.
@@ -554,7 +539,7 @@ class SurvivalHarmonium(BaseHarmonium):
         P = self.p_x_condition_h_binary(H)
 
         # Turn hidden unit on when probability is larger than random uniform number,
-        X = (P > U).astype(float)
+        X = (P > U).astype(float64)
         return X
 
     def sample_x_gamma(self, H: Matrix) -> Matrix:
@@ -633,12 +618,28 @@ class SurvivalHarmonium(BaseHarmonium):
         """
         Verify that input data adheres to the bounds.
         """
-        xi_A = self._get_columns(X, self.categorical_columns)
+        xi_A = get_columns(X, self.categorical_columns)
         if xi_A.size > 0:
             distinct_categories = unique(xi_A[~isnan(xi_A)])
-            if set(distinct_categories) != {0, 1}:
+            if len(distinct_categories) > 2:
+                # Allow floating point values for categories, but only if on the
+                # unit interval.
                 if any((xi_A < 0) | (xi_A > 1)):
                     raise ValueError("Categorical data not in the [0,1] interval.")
+            elif not set(distinct_categories.astype(int)).issubset({0, 1}):
+                raise ValueError("Categorical data not in the [0,1] interval.")
+
+        events = get_columns(X, self.event_columns)
+        if events.size > 0:
+            indicators = unique(events)
+            if len(indicators) > 2:
+                raise ValueError(
+                    f"Invalid event indicator values in columns {self.event_columns}."
+                )
+            elif not set(indicators.astype(int)).issubset({0, 1}):
+                raise ValueError(
+                    f"Indicator values (in columns {self.event_columns}) must be 0 or 1."
+                )
 
     def fit(self, X: Matrix, y=None):
         """
@@ -647,6 +648,7 @@ class SurvivalHarmonium(BaseHarmonium):
         self.is_fitted_ = False
         self.n_hidden_units = int(self.n_hidden_units)
         self.check_X(X)
+        X = X.astype(float64)
 
         self.weight_decay_ = self._parse_weight_decay()
         self.eps = self._parse_learning_rate()
@@ -654,7 +656,7 @@ class SurvivalHarmonium(BaseHarmonium):
         self._parse_column_names(X)
 
         # Determine time horizon.
-        XI_B = self._get_columns(X, self.survival_columns)
+        XI_B = get_columns(X, self.survival_columns)
         self.time_horizon_ = array([1.0 for _ in range(len(self.survival_columns))])
         if self.time_horizon is not None and XI_B.size > 0:
             if isinstance(self.time_horizon, float):
@@ -669,7 +671,7 @@ class SurvivalHarmonium(BaseHarmonium):
         self.validation_metrics_: Dict[str, list] = defaultdict(list)
 
         # Set random state.
-        np.random.seed(self.random_state)
+        reset_random_state(self.random_state)
 
         if self.survival_columns:
             # All the time points are the same when a single float is passed.
@@ -682,8 +684,8 @@ class SurvivalHarmonium(BaseHarmonium):
                 t_median = []
                 i_col = 0
                 for t_col, e_col in zip(self.survival_columns, self.event_columns):
-                    t = self._get_columns(X, [t_col]).flatten()
-                    e = self._get_columns(X, [e_col]).flatten()
+                    t = get_columns(X, [t_col]).flatten()
+                    e = get_columns(X, [e_col]).flatten()
                     kmf = KaplanMeierFitter().fit(t, event_observed=e)
                     t_med = (
                         kmf.median_survival_time_ / self.time_horizon_.flatten()[i_col]
@@ -912,7 +914,7 @@ class SurvivalHarmonium(BaseHarmonium):
 
         # Gamma harmonium: -ln [Gamma(a) gamma*(a,b)].
         ALPHA, BETA = self.alpha_beta(H)
-        F_B = -log(normalisation_gamma_distribution(ALPHA, BETA)).sum(axis=1)
+        F_B = -log_normalisation_gamma_distribution(ALPHA, BETA).sum(axis=1)
 
         # Gaussian harmonium: <a/sigma|wh> - 1/2|Wh|^2 - ln[2pi sigma^2].
         F_C = (
@@ -932,8 +934,27 @@ class SurvivalHarmonium(BaseHarmonium):
         """
         Unpack and inflate variables and events in triplets.
         """
-        xi, event = self._unpack(X)
-        return (self.fill_nan(xi), self.inflate_event_masks(xi, event))
+        (xi_A, xi_B, xi_C), event_B = self._unpack(X, verify=False)
+        event_A = event_C = array([], dtype=bool_)
+
+        nan_A = isnan(xi_A, order="C", dtype=bool_)
+        nan_C = isnan(xi_C, order="C", dtype=bool_)
+        # Use the column indices [i.e., where(mask)[1]] to pick the
+        # corresponding fillers.
+        if nan_A.any():
+            fillers_A = get_columns(self.fill_nan_, self.categorical_columns)
+            xi_A[nan_A] = take(fillers_A, where(nan_A)[1])
+            event_A = ~nan_A
+        if nan_C.any():
+            fillers_C = get_columns(self.fill_nan_, self.numeric_columns)
+            xi_C[nan_C] = take(fillers_C, where(nan_C)[1])
+            event_C = ~nan_C
+
+        # For the time-to-event variables (group B), replace nan->0.
+        # This ensures that `sample_x_wake` generates samples from
+        # the entire [0,1] interval (instead of [xi, 1]).
+        xi_B[isnan(xi_B)] = 0.0
+        return ((xi_A, xi_B, xi_C), (event_A, event_B, event_C))
 
     def inflate_event_masks(
         self, X: MatrixTriplet, event: Matrix = array([])
@@ -1065,7 +1086,12 @@ class SurvivalHarmonium(BaseHarmonium):
             boundaries = append(boundaries_B, boundaries_C).reshape(-1, 2)
 
             integral_function = self._lambda_partition_function_i(
-                x_A, x_B, x_C, event_A, event_B, event_C,
+                x_A,
+                x_B,
+                x_C,
+                event_A,
+                event_B,
+                event_C,
             )
             value, _ = nquad(integral_function, boundaries)
             Z[k] = value
@@ -1149,7 +1175,9 @@ class SurvivalHarmonium(BaseHarmonium):
         return S
 
     def _proba_singletons(
-        self, X: Matrix, survival_distribution: bool = True,
+        self,
+        X: Matrix,
+        survival_distribution: bool = True,
     ) -> Matrix:
         r"""
         Discriminative distribution per event variable x_i.
@@ -1191,7 +1219,7 @@ class SurvivalHarmonium(BaseHarmonium):
         """
 
         def unnormalised_time(column):
-            """ Compute unnormalised risk_score_time_point time. """
+            """Compute unnormalised risk_score_time_point time."""
             i = self.survival_columns.index(column)
             return self.risk_score_time_point_[i] * self.time_horizon_.flatten()[i]
 
@@ -1342,11 +1370,11 @@ class SurvivalHarmonium(BaseHarmonium):
             i = self.survival_columns.index(time_column)
             event_column = self.event_columns[i]
             brier_i = brier_loss(
-                train_time=self.X_train[time_column],
-                train_event=self.X_train[event_column],
-                test_time=X[time_column],
-                test_event=X[event_column],
-                S_pred=predicted_risk[time_column],
+                train_time=get_columns(self.X_train, time_column),
+                train_event=get_columns(self.X_train, event_column),
+                test_time=get_columns(X, time_column),
+                test_event=get_columns(X, event_column),
+                S_pred=predicted_risk[time_column].to_numpy(),
                 tau=tau,
             )
             brier_scores[time_column] = brier_i
@@ -1356,6 +1384,33 @@ class SurvivalHarmonium(BaseHarmonium):
 
         return brier_scores
 
+    def bayesian_information_criterion(self, X: Matrix, negate: bool = False):
+        """
+        Bayesian information criterion for censored data.
+        """
+        likelihood = exp(self.log_likelihood(X).sum())
+        event = get_columns(X, self.event_columns, dtype=bool_)
+        # Instead of number of samples, we use the number of events that are
+        # completely observed.
+        d = (event.sum(axis=1) == self.n_event_units).sum()
+        # Number of free parameters.
+        k = (
+            self.W_A.size
+            + self.a_A.size
+            + self.W_B.size
+            + self.V.size
+            + self.a_B.size
+            + self.c.size
+            + self.W_C.size
+            + self.a_C.size
+            + self.sigma.size
+            + self.b.size
+        )
+        sgn = 1
+        if negate:
+            sgn = -1
+        return sgn * (k * log(d) - 2 * likelihood)
+
     def concordance_index(
         self,
         X: Matrix,
@@ -1561,15 +1616,34 @@ class SurvivalHarmonium(BaseHarmonium):
         for h in generate_binary_permutations(self.n_hidden_units):
             # Do the sum over `h` for all `m` records in X. => Repeat `h` m times.
             H = h.reshape(1, -1).repeat(m, axis=0)
-            total_product = self.energy_factor_binary(X_A, H, event_A)
-            total_product *= self.energy_factor_gamma(X_B, H, event_B)
-            total_product *= self.energy_factor_gauss(X_C, H, event_C)
-
+            # Compute product as log sums for numerical stability.
+            log_sum = self.log_energy_factor_binary(X_A, H, event_A)
+            log_sum += self.log_energy_factor_gamma(X_B, H, event_B)
+            log_sum += self.log_energy_factor_gauss(X_C, H, event_C)
+            log_sum += (-H @ self.b).flatten()
+            total_product = exp(log_sum)
             if np.any(total_product < 0):
                 raise ValueError
-            Z += total_product * exp(-H @ self.b).flatten()
+            Z += total_product
         return Z
 
+    def log_energy_factor_binary(
+        self, X: Matrix, H: Matrix, mask: Matrix
+    ) -> np.ndarray:
+        """
+        Missing data corrected energy contribution of the categoric variables (group A).
+        """
+        if mask.size == 0:
+            return -self.energy_binary(X, H)
+
+        p = zeros_like(X)
+        z = self.z(H)
+        # Observed values.
+        p[mask] = -(X * z)[mask]
+        # Marginalise out missing values.
+        p[~mask] = logaddexp(-z[~mask], 0)
+        return p.sum(axis=1)
+
     def energy_factor_binary(self, X: Matrix, H: Matrix, mask: Matrix) -> np.ndarray:
         """
         Missing data corrected energy contribution of the categoric variables (group A).
@@ -1585,6 +1659,28 @@ class SurvivalHarmonium(BaseHarmonium):
         p[~mask] = 1 + exp(-z[~mask])
         return p.prod(axis=1)
 
+    def log_energy_factor_gauss(self, X: Matrix, H: Matrix, mask: Matrix) -> np.ndarray:
+        """
+        Missing data corrected energy contribution of the numeric variables (group C).
+        """
+        if mask.size == 0:
+            return -self.energy_gauss(X, H)
+
+        p = zeros_like(X)
+        # 1)
+        # Observed variables.
+        energy = self.energy_gauss(X, H, as_matrix=True)
+        p[mask] = -energy[mask]
+
+        # 2)
+        # Missing values are integrated out.
+        linear = self.a_C.T / self.sigma.T * (H @ self.W_C.T)
+        quadratic = -((H @ self.W_C.T) ** 2) / 2
+        phase = linear + quadratic
+        p[~mask] = (0.5 * log(2 * pi * self.sigma ** 2).flatten() - phase)[~mask]
+
+        return p.sum(axis=1)
+
     def energy_factor_gauss(self, X: Matrix, H: Matrix, mask: Matrix) -> np.ndarray:
         """
         Missing data corrected energy contribution of the numeric variables (group C).
@@ -1631,3 +1727,29 @@ class SurvivalHarmonium(BaseHarmonium):
         )
 
         return p.prod(axis=1)
+
+    def log_energy_factor_gamma(self, X: Matrix, H: Matrix, mask: Matrix) -> np.ndarray:
+        r"""
+        Censor corrected energy factor of the time-to-event variables (group B).
+
+        Calculates:
+        -\tilde{E}_B(o, h) = log[\int dx^B \exp[-E_B(x^B, h)] \chi(x^B, o^B)]
+        """
+        if mask.size == 0:
+            return -self.energy_gamma(X, H)
+
+        p = zeros_like(X)
+        # 1)
+        # Observed values.
+        energy = self.energy_gamma(X, H, as_matrix=True)
+        p[mask] = -energy[mask]
+        # 2)
+        # For censored values, integrate out the unobserved region (the interval
+        # from censor time to 1).
+        alpha, beta = self.alpha_beta(H)
+        p_surv = normalisation_gamma_interval_distribution(
+            alpha[~mask], beta[~mask], t_left=X[~mask], t_right=1.0
+        )
+        p[~mask] = log(p_surv, out=full_like(p_surv, -inf), where=(p_surv != 0))
+
+        return p.sum(axis=1)
